'''
TransMap
start date: 13.12.18
author: Nikhil VJ
Created for work on TSRTC Hyderabad bus routes
'''

print('\nTransMap')
print('Starting up the program, loading dependencies, please wait...\n\n')

import datetime
import json
import os
import signal  # for catching Ctrl+C and exiting gracefully.
import sys
import time, datetime
import webbrowser
from collections import OrderedDict

import pandas as pd
import tornado.ioloop
import tornado.web
import xmltodict

# setting constants
root = os.path.dirname(__file__) # needed for tornado
routesFolder = 'routes/'
timeFolder = 'start_times/'
# dataFolders = ['data/','done/', 'hcb_com_routes/']
csvFolder = 'csv_routes/'

for folder in ([routesFolder, timeFolder, csvFolder]):
	if not os.path.exists(folder):
		print('Creating folder {} in program path'.format(folder))
		os.makedirs(folder)

print('Loaded dependencies, starting program.')

###########################
# Define all API call classes here

class loadRoutesList(tornado.web.RequestHandler):
	def get(self):
		# /API/loadRoutesList
		start = time.time()

		returnContent = '<option value="">Select a route</option>'

		# LOOP to first list csv's, then 
		for f in os.listdir(csvFolder):
			if f.lower().endswith('.csv') and os.path.isfile(csvFolder + f):
				returnContent += '<option value="{}">{}</option>'.format((csvFolder+f),f)
		
		# LOOP to list subfolders as groups and csv's inside them
		subfoldersList = [f for f in os.listdir(csvFolder) if os.path.isdir(csvFolder + f)]
		subfoldersList.sort() # having to sort this because herokuapp's filesystem is crazy
		print('sorted subfolders:', ', '.join(subfoldersList) )

		for f in subfoldersList:
			#if os.path.isdir(csvFolder + f):
			subFolder = csvFolder + f + '/'
			returnContent+= '<optgroup label="{}">'.format(f)
			
			csvList = [f2 for f2 in os.listdir(subFolder) if ( f2.lower().endswith('.csv') and os.path.isfile(subFolder + f2) )]
			csvList.sort()
			print('sorted csvs in subfolder',subFolder,':')
			print( ', '.join(csvList[0:5]),'...')
			
			for f2 in csvList:
				#if f2.lower().endswith('.csv') and os.path.isfile(subFolder + f2):
				returnContent += '<option value="{}">{}</option>'.format(subFolder+f2, f2)
			returnContent += '</optgroup>'
		self.write(returnContent)
		end = time.time()
		print("loadRoutesList GET call took {} seconds.".format(round(end-start,2)))

class loadRouteCSV(tornado.web.RequestHandler):
	def get(self):
		# /API/loadRouteCSV?route=filename
		#time.sleep(1)
		start = time.time()
		filename = self.get_argument('route',default='')
		# print(filename)
		try:
			df = pd.read_csv(filename, dtype=str).fillna('')
			# print(df.head())
		except:
			df = pd.DataFrame()
		
		csvName = filename.split('/')[-1]
		# print('timings:',timeFolder + csvName)

		try:
			dfTime = pd.read_csv(timeFolder + csvName, dtype=str).fillna('')
			# print(dfTime.head())
		except:
			dfTime = pd.DataFrame()
		
		data = OrderedDict({ 'filename': csvName[0:-4] })

		if 'route' in df.columns:
			data['route'] = df['route'].iloc[0]
		else: data['route'] = ''

		data['stops0'] = '\n'.join( df[df.direction == '0']['stop_name'].tolist() )
		data['stops1'] = '\n'.join( df[df.direction == '1']['stop_name'].tolist() )
		if 'direction' in dfTime.columns:
			data['timings0'] = '\n'.join( dfTime[dfTime.direction == '0']['start_time'].apply(lambda x: x[0:5]).tolist() )
			data['timings1'] = '\n'.join( dfTime[dfTime.direction == '1']['start_time'].apply(lambda x: x[0:5]).tolist() )
		
		# print(data)

		self.write(json.dumps(data))
		end = time.time()
		print("loadRouteCSV GET call took {} seconds.".format(round(end-start,2)))


class saveRoute(tornado.web.RequestHandler):
	def post(self):
		# /API/saveRoute
		start = time.time()
		data = json.loads( self.request.body.decode('UTF-8') )
		# print(data)
		# since incoming data is with the whole json thingy anyway, might as well just dump it to a json file
		filename = data.get('routeFileName','')
		if not filename : filename = 'dump'
		# in case the route filename isn't there, save as 'dump.json'
		
		# with open(routesFolder + filename +'.json','w') as f:
		#	f.write(self.request.body.decode('UTF-8'))
		
		# 8.12.18: intervention: re-process the stops to make a json array
		data['stopsArray0'] = []
		data['stopsArray1'] = []
		for stop in data.get('stops0','').split('\n'):
			if stop != '':
				stopRow = {'stop_name': stop}
				data['stopsArray0'].append(stopRow)
		
		for stop in data.get('stops1','').split('\n'):
			if stop != '':
				stopRow = {'stop_name': stop}
				data['stopsArray1'].append(stopRow)
		
				
		json.dump(data, open(routesFolder + filename +'.json','w'), indent='\t', ensure_ascii=True)
		
		self.write('Ok got it bro.')
		end = time.time()
		print("saveRoute POST call took {} seconds.".format(round(end-start,2)))


class loadJsonsList(tornado.web.RequestHandler):
	def get(self):
		# /API/loadJsonsList
		start = time.time()
		returnContent = recursiveDropdown(routesFolder,ext='.json')
		# filenames = [ f for f in os.listdir(routesFolder) if f.lower().endswith('.json') ]
		# filenames.sort()
		# self.write(','.join(filenames))
		self.write(returnContent)
		end = time.time()
		print("loadJsonsList GET call took {} seconds.".format(round(end-start,2)))


class loadJson(tornado.web.RequestHandler):
	def get(self):
		# /API/loadJson?route=filename
		start = time.time()
		filename = self.get_argument('route',default='')
		if not filename or not os.path.exists(routesFolder + filename):
			self.set_status(400)
			self.write("Error: invalid filename.")
			return
		
		with open(routesFolder + filename) as fh:
			a = json.load(fh)
		self.write(json.dumps(a) )
		end = time.time()
		print("loadJson GET call took {} seconds.".format(round(end-start,2)))

###################
# List API calls and the class they will be directed to, here
def make_app():
	return tornado.web.Application([
		#(r"/API/data", APIHandler),
		(r"/API/loadRoutesList", loadRoutesList),
		(r"/API/loadRouteCSV", loadRouteCSV),
		(r"/API/saveRoute", saveRoute),
		(r"/API/loadJsonsList", loadJsonsList),
		(r"/API/loadJson", loadJson),
		(r"/(.*)", tornado.web.StaticFileHandler, {"path": root, "default_filename": "index.html"})
	])


###################
# Done with API calls defining. 

####################
# Functions

# for catching Ctrl+C and exiting gracefully. From https://nattster.wordpress.com/2013/06/05/catch-kill-signal-in-python/
def signal_term_handler(signal, frame):
	# to do: Make this work in windows, ra!
	print('\nClosing Program.\nThank you for using Routes Info Collector.\n')
	sys.exit(0)

def recursiveDropdown(targetFolder = '.', ext=''):
	'''
	recursiveDropdown(targetFolder = '.', ext='')
	recursively scans targetFolder for all files ending with extension ext (or all files by default), and generates innerHTML for a dropdown menu, with respective sub-folder paths as optgroups.
	'''
	if targetFolder.endswith('/'): targetFolder = targetFolder[:-1]
	content = '<option value="">Select one</option>'

	# traverse through all files, folders under a path recursively. from https://stackoverflow.com/a/2212698/4355695
	for root, subdirs, files in os.walk(targetFolder):
		folder = root[len(targetFolder)+1:]
		targetFiles = [x for x in files if x.endswith(ext)]
		targetFiles.sort()
		
		if len(folder) and len(targetFiles):
			content += '<optgroup label="{}">'.format(folder)
		for file in targetFiles:
			filepath = os.path.join(root,file)[len(targetFolder)+1:]
			content += '<option value="{}">{}</option>'.format(filepath,file)
		if len(folder) and len(targetFiles):
			content += '</optgroup>'
	return content

###################################################			
# the __main__ function sets up the tornado server

if __name__ == "__main__":
	signal.signal(signal.SIGINT, signal_term_handler)
	app = make_app()
	portnum = 5000
	while True: # loop to increment the port number till we find one that isn't occupied
		try:
			port = int(os.environ.get("PORT", portnum))
			app.listen(port)
			break
		except OSError:
			portnum += 1
			if portnum > 9999: 
				print('Thats it I give up! From 5000 to 9999 not a single port number is free! Can\'t run your server, sorry!')
				sys.exit()

	thisURL = "http://localhost:" + str(port)
	webbrowser.open(thisURL)
	print("Open " + thisURL + " in your Web Browser if you don't see it opening automatically within 5 seconds.")
	tornado.ioloop.IOLoop.current().start()
