'''
RouteMapper
date: 21.10.18
author: Nikhil VJ
formerly named FuzzyMapper
'''
print('\nRoute Mapper')
print('Starting up the program, loading dependencies, please wait...\n\n')

import tornado.web
import tornado.ioloop
import json
import os
import time, datetime
import pandas as pd
from collections import OrderedDict
import webbrowser
import signal, sys # for catching Ctrl+C and exiting gracefully.

import jellyfish as jf
from fuzzywuzzy import fuzz
'''
import gpxpy
import gpxpy.gpx
import io
'''
import xmltodict

# setting constants
root = os.path.dirname(__file__) # needed for tornado
dataFolder = 'data/'
timeFolder = 'start_times/'
backupFolder = 'data/backups/'
databankFolder = 'databank/'
configFile = 'config.json'
# 23.9.18: set the actual databank file to be loaded dynamically by reading the .json when the API call happens.
# look for "databank" below.

# create backup folder if it doesn't exist
if not os.path.exists(dataFolder):
	os.makedirs(dataFolder)
if not os.path.exists(backupFolder):
	os.makedirs(backupFolder)

print('Loaded dependencies, starting program.')

###########################
# Define all API call classes here

class left(tornado.web.RequestHandler):
	def get(self):
		# /API/left?filename=filename
		start = time.time()
		filename = self.get_argument('filename',default='')
		print(filename)

		if os.path.exists(dataFolder + filename):
			leftDict = pd.read_csv(dataFolder+filename, dtype=str).fillna('').to_dict(orient='records', into=OrderedDict)
		else:	
			self.set_status(400)
			self.write('File not found')
			return
	
		if os.path.exists(timeFolder + filename):
			try:
				timingsDF = pd.read_csv(timeFolder + filename, dtype=str).fillna('').drop_duplicates(['direction','start_time'])
				timings0 = timingsDF[timingsDF['direction'].isin(['0','',0]) ]['start_time'].tolist()
				timings1 = timingsDF[timingsDF['direction'].isin(['1',1]) ]['start_time'].tolist()
			except (OSError, KeyError) as e:
				print('Error reading',timeFolder + filename,e, 'so skipping reading timings')
				timings0 = []
				timings1 = []
		else: 
			timings0 = []
			timings1 = []

		returnDict = { 'sequence': leftDict, 'timings0':timings0, 'timings1':timings1  }
		
		self.write(json.dumps(returnDict))

		end = time.time()
		print("left GET call took {} seconds.".format(round(end-start,2)))

	def post(self):
		# /API/left?filename=filename
		start = time.time()	
		filename = self.get_argument('filename',default='')
		data = json.loads( self.request.body.decode('UTF-8') )
		
		if os.path.exists(filename):
			timestamp = '{:%Y%m%d.%H%M%S}'.format(datetime.datetime.now())
			os.rename(filename, 'backups/{}.backup-{}.csv'.format(filename,timestamp))
		
		df = pd.DataFrame( data.get('sequence',[]) )
		if 'stop_desc' in df.columns:
			df = df[ df.stop_desc != 'BAD']
		df.to_csv(dataFolder+filename, index=False)
		print('Saved',dataFolder+filename,'with',len(df),'entries')
		del df

		saveTimings( filename, data.get('timings0',[]), data.get('timings1',[]) )

		self.write('Saved route information to CSVs.')
		end = time.time()
		print("left POST call took {} seconds.".format(round(end-start,2)))

class right(tornado.web.RequestHandler):
	def get(self):
		# /API/right?searchName=${searchName}&fuzzyFlag=${fuzzyFlag}&jellyFlag=${jellyFlag}
		start = time.time()
		searchName = self.get_argument('searchName',default='')
		print(searchName)
		fuzzyFlag = self.get_argument('fuzzyFlag',default='n')
		jellyFlag = self.get_argument('jellyFlag',default='n')

		with open(configFile) as f:
			configDict = json.load(f)
		databank = configDict.get('databank', 'hydstopsGoogle3.csv')

		try:
			df = pd.read_csv(databankFolder+databank)
		except OSError as e:
			print('Error:',e)
			self.set_status(400)
			self.write('Error:' + str(e))
			return

		if jellyFlag == 'y':
			Jstart = time.time()
			df2 = jellyPanda(searchName,df)
			Jend = time.time()
			print("Jellyfish scoring took {} seconds.".format(round(Jend-Jstart,3)))	
		else: df2 = df
		
		if fuzzyFlag == 'y':
			Fstart = time.time()
			df3 = fuzzyPanda(searchName,df2)
			Fend = time.time()
			print("Fuzywuzzy scoring took {} seconds.".format(round(Fend-Fstart,3)))	
		else: df3 = df2
		
		rightJson = df3.to_json(orient='records', force_ascii=False)
		
		del df,df2,df3 # cleaning up
		
		self.write(rightJson)
		end = time.time()
		print("right GET call took {} seconds.".format(round(end-start,2)))

class mapExact(tornado.web.RequestHandler):
	def get(self):
		# /API/mapExact?filename=filename
		start = time.time()
		filename = self.get_argument('filename',default='')
		print(filename)
		
		with open(configFile) as f:
			configDict = json.load(f)
		databank = configDict.get('databank', 'hydstopsGoogle3.csv')

		try:
			df = pd.read_csv(dataFolder+filename, dtype=str).fillna('')
			dfBank = pd.read_csv(databankFolder+databank, dtype=str).fillna('')
		except OSError as e:
			print('Error:',e)
			self.set_status(400)
			self.write('Error:' + str(e))
			return

		arr, count = mapExactFunc(df,dfBank)
		
		pd.DataFrame(arr).to_csv(dataFolder+filename, index=False)
		
		self.write('Mapped {} exact matches.'.format(count) )
		end = time.time()
		print("mapExact GET call took {} seconds.".format(round(end-start,2)))

class GPXUpload(tornado.web.RequestHandler):
	def post(self):
		# /API/GPXUpload
		start = time.time()

		gpx_string = self.request.files['gpx'][0]['body']
		gpxDict = xmltodict.parse(gpx_string, attr_prefix='')
		
		waypointDF1 = pd.DataFrame(gpxDict['gpx']['wpt'])
		waypointDF2 = waypointDF1[waypointDF1.name.str.lower() != 'bus stop'][['name','lat','lon','time']].copy()
		# removing extra bus stop entries
		waypointDict = waypointDF2.to_dict(orient='records', into=OrderedDict)
		
		trackDict = [ [ x['lat'], x['lon'] ] for x in gpxDict['gpx']['trk']['trkseg']['trkpt'] ]

		returnJson = json.dumps({'waypoint':waypointDict, 'track':trackDict}, ensure_ascii=False)
		self.write(returnJson)
		end = time.time()
		print("GPXUpload POST call took {} seconds.".format(round(end-start,2)))


class getDataFiles(tornado.web.RequestHandler):
	def get(self):
		# /API/getDataFiles?filename=filename
		start = time.time()
		
		filenames = [f for f in os.listdir(dataFolder) 
			if f.lower().endswith('.csv') 
			and os.path.isfile(os.path.join(dataFolder, f))]
		
		filenames.sort()
		content = ''
		for csv in filenames:
			content += '<option value="{}">{}</option>'.format(csv,csv)

		self.write(content)
		end = time.time()
		print("getDataFiles GET call took {} seconds.".format(round(end-start,2)))

###################
# List API calls and the class they will be directed to, here
def make_app():
	return tornado.web.Application([
		#(r"/API/data", APIHandler),
		(r"/API/left", left),
		(r"/API/right", right),
		(r"/API/mapExact", mapExact),
		(r"/API/GPXUpload", GPXUpload),
		(r"/API/getDataFiles", getDataFiles),
		(r"/(.*)", tornado.web.StaticFileHandler, {"path": root, "default_filename": "index.html"})
	])


###################
# Done with API calls defining. 

####################
# Functions


# for catching Ctrl+C and exiting gracefully. From https://nattster.wordpress.com/2013/06/05/catch-kill-signal-in-python/
def signal_term_handler(signal, frame):
	# to do: Make this work in windows, ra!
	print('\nClosing Program.\nThank you for using Route Mapper.\n')
	sys.exit(0)


zapper = lambda x: ''.join(e.lower() for e in x if e.isalnum())
# lowercase, and zap all chars that aren't a letter or digit. ex: A.S. Colony >> ascolony
# from https://stackoverflow.com/a/5843560/4355695
# lambda function : declared as a variable. from http://www.diveintopython.net/power_of_introspection/lambda_functions.html


def jellyPanda(searchName,df):
	# to do : make everything lowercase before comparing. Actually, check if any difference.
	# shit, it does make a difference. Lowercase everything and then compare!
	searchName = zapper(searchName)

	df['Jlevenshtein'] = df.stop_name.apply( lambda x: jf.levenshtein_distance(searchName,zapper(x)) )
	df['Jdamerau'] = df.stop_name.apply( lambda x: jf.damerau_levenshtein_distance(searchName,zapper(x)) )
	df['Jhamming'] = df.stop_name.apply( lambda x: jf.hamming_distance(searchName,zapper(x)) )
	df['JjaroD'] = df.stop_name.apply( lambda x: jf.jaro_distance(searchName,zapper(x)) )
	df['JjaroW'] = df.stop_name.apply( lambda x: jf.jaro_winkler(searchName,zapper(x)) )
	df['Jmatch'] = df.stop_name.apply( lambda x: jf.match_rating_comparison(searchName,zapper(x)) )
	return df


def fuzzyPanda(searchName, df, nameColumn='stop_name'):
	searchName = zapper(searchName)

	df['Fsimple'] = df[nameColumn].apply( lambda x: fuzz.ratio(searchName,zapper(x)) )
	df['Fpartial'] = df[nameColumn].apply( lambda x: fuzz.partial_ratio(searchName,zapper(x)) )
	df['FtokenSort'] = df[nameColumn].apply( lambda x: fuzz.token_sort_ratio(searchName,zapper(x)) )
	df['FtokenSet'] = df[nameColumn].apply( lambda x: fuzz.token_set_ratio(searchName,zapper(x)) )
	return df


def mapExactFunc(df,dfBank,\
	matchColumns=['stop_name','stop_name'],
	checkFilled =  ['stop_lat','stop_lon'],
	# copyColumns=['stop_lat','stop_lon','stop_id:source_stop_id','source','stop_name:source_stop_name','stop_desc'] ):
	# copyColumns=['stop_lat','stop_lon','source','stop_name:source_stop_name','stop_desc'] ):
	copyColumns=['source','stop_name:source_stop_name','stop_desc'] ):
	'''
	Function for bulk mapping from data bank if there are identical matches
	plan:
	- convert df to list-of-dicts (records)
	- loop through each dict
	- process record if it hasn't been mapped yet (lat, long empty)
	- check if name is there in databank. take first match if multiple.
	- copy over copyColumns, and use source:destination to rename the copied column
	- return list of dicts
	to do:
	- sanity check: try next match if this one's lat longs are too far away.
	'''
	
	df['zapped'] = df[matchColumns[0] ].apply( zapper )
	dfBank['zapped'] = dfBank[matchColumns[1] ].apply( lambda x: x.lower().replace('bus stop','') )
	dfBank['zapped'] = dfBank[matchColumns[1] ].apply( zapper )
	print('df:')
	print(df.head())
	print('dfBank:')
	print(dfBank.head())
	
	arr = df.to_dict(orient='records', into=OrderedDict)
	count = 0
	foundList = []
	for i,row in enumerate(arr):

		# process only those rows that do NOT have lat,lon already mapped
		if not (row.get('stop_lat',None) and row.get('stop_lon',None) ): 
			# to do: parameterized this to checkFilled list -- NAH lets keep lat, lon as special
			matches = dfBank[ dfBank['zapped'] == row['zapped'] ]
			if len(matches):

				# if(i%100 == 1):
				#	print('\n\nrow in to-map table:',row)
				#	print('Matching databank table:')
				#	print(matches)

				row['stop_lat'] = matches.iloc[0]['stop_lat']
				row['stop_lon'] = matches.iloc[0]['stop_lon']

				# copying remaining metadata
				for col in copyColumns:
					# check if it's one column name identical in both tables, or a source:destination pair
					if ':' in col:
						source,destination = col.split(':')
						row[destination] = matches.iloc[0][source]
					else:
						row[col] = matches.iloc[0][col] # take the first match only
				count +=1
				foundList.append(row[matchColumns[0]])
		row.pop('zapped',None) # was going to get rid of this, but decided to keep it.

	print('Matches found for {} names.\n'.format(count) )
	# print(foundList)
	return arr, count

def saveTimings( filename, timings0, timings1 ):
	collector = []
	for (timings, direction) in [(timings0,'0'), (timings1,'1')]:
		for timing in timings:
			if not validateTiming(timing): continue #silently skip to next value
			row = OrderedDict({ 'route': filename[:-4], 'direction':direction, 'start_time':timing })
			collector.append(row)
	pd.DataFrame(collector).to_csv(timeFolder + filename, index_label='sr')
	print('Saved {} with {} trip start times.'.format(timeFolder + filename, len(collector)))
	del collector

def validateTiming(a):
	if isinstance(a, str):
		if len(a) == 8 and a.replace(':','',2).isdigit():
			return True
	return False

###################################################			
# the __main__ function sets up the tornado server

if __name__ == "__main__":
	signal.signal(signal.SIGINT, signal_term_handler)
	app = make_app()
	portnum = 5000
	while True: # loop to increment the port number till we find one that isn't occupied
		try:
			port = int(os.environ.get("PORT", portnum))
			app.listen(port)
			break
		except OSError:
			portnum += 1
			if portnum > 9999: 
				print('Thats it I give up! From 5000 to 9999 not a single port number is free! Can\'t run your server, sorry!')
				sys.exit()

	thisURL = "http://localhost:" + str(port)
	webbrowser.open(thisURL)
	print("Open " + thisURL + " in your Web Browser if you don't see it opening automatically within 5 seconds.")
	tornado.ioloop.IOLoop.current().start()


